/* $Id: http-rman.c,v 1.2 1994/05/15 14:55:12 fredrik Exp $
 *
 * Name:
 *	http-rman.c -- a rudimentary man-page HTTP server
 *
 * Description:
 *	This is a minimal HTTP server using RosettaMan by T.A. Phelps
 *	(phelps@ACM.org) to produce hypertext renditions
 *	of man-pages on the fly.
 *
 *	This server processes URLs with the following syntax:
 *
 *		[/man] ?<command> [ ?<section> ]
 *
 *	For URLs matching this format, it pipes the output of
 *	man <section> <command> through rman and sends it to
 *	the HTTP client. For other URLs, it returns the document
 *	given as argv[1] (using cat(1). The leading /man is
 *	optional, but is strongly recommended.
 *
 *	This server is shipped as two files, the http-rman.c
 *	sources and the http-rman.html sample frontpage. I have
 *	not included a Makefile; you can write your own or just
 *	type [g]cc -o http-rman http-rman.c
 *
 * What do I need to run this:
 *	If you don't have it, pick up RosettaMan by anonymous ftp
 *	from ftp.cs.berkeley.edu: /ucb/people/phelps/tcl/rman.tar.Z
 *
 *	You'll also need an HTTP client such as NCSA Mosaic to talk
 *	to this server. Mosaic is available by anonymous FTP from
 *	ftp://ftp.ncsa.uiuc.edu/Mosaic 
 *
 *	Both RosettaMan (rman) and Mosaic are available from many
 *	other sites. Try Archie, or check your local or national net
 *	archive.
 *
 * How do I get it running:
 *	First, compile the server (see above), and install it
 *	somewhere.
 *
 *	The server runs under inetd(8). Add a service to
 *	/etc/services, say:
 *
 *		http-rman 4080/tcp
 *
 *	If you're not about to install another HTTP server on your
 *	machine, you may use the default HTTP port, 80, instead.
 *
 *	Then add an entry to /etc/inetd.conf, such as (on a single line):
 *
 *		http-rman stream tcp nowait root /usr/local/bin/http-rman
 *			http-rman /usr/local/lib/rman/http-rman.html
 *
 *	Change /usr/local/bin and /usr/local/lib/rman to where you
 *	installed the two files. In addition, you may wish to run
 *	the server as something other than root...
 *
 *	Restart inetd(8) (use kill -HUP or kill it and start it again)
 *	and try the following:
 *
 *	$ Mosaic http://localhost:4080
 *
 *	If you don't have Mosaic, try the following instead:
 *
 *	$ telnet localhost 4080
 *	Trying 127.0.0.1...
 *	Connected to localhost.
 *	Escape character is '^]'.
 *	GET /man?ls <return>
 *	<return>
 *	HTTP/1.0 200 OK
 *	...
 *
 * Portability:
 *	You'll need an ANSI compiler (or an editor and some patience).
 *	As it stands right now, this code has been successfully
 *	compiled on OSF/1 AXP using cc, and on SunOS 4.1 using gcc.
 *	Might need some tuning for other platforms.
 *
 * Legal Issues:
 *	Check the external visibility of the http-rman service
 *	you choose. This server gives a user access to ALL man-
 *	pages on your machine. You may have installed copyrighted
 *	software (your operating system, for example) with
 *	man-pages that you are NOT allowed to make visible for
 *	anyone out there...
 *
 * History:
 *	94-04-30 fl: created
 *	94-05-13 fl: stripped away everything but rman support
 *      
 *      96-05-03 Roland Dyroff dyroff@suse.de:
 *               hacked a lot to fit suse's needs
 *      98-02-17 hacked a lotter :-) to make it a bit more robust
 *               and "secure" tmg@suse.de
 *	99-10-29 More secure: use mkstemp and remove temp file
 *		 even if signaled.  werner@suse.de
 *	00-01-26 Much more secure: dont allow "-" as first character
 *              and print exploit attempts to syslog by default. marc@suse.de
 *
 * Copyright (c) Fredrik Lundh 1994 (fredrik_lundh@ivab.se)
 * All rights reserved.
 *
 * Copyright  (c)  1996-2000  SuSE GmbH Nrnberg, Germany.
 *
 * please send bugfixes or comments to feedback@suse.de.
 */


#include <stdio.h>		/* printf(), fflush(stdio) etc */
#include <string.h>		/* strrchr(), strcmp() etc */
#include <syslog.h>
#include <ctype.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <errno.h>
#include <signal.h>

/* PATH to man command */
#define MAN_PATH "/usr/bin/man"
/* path to rosetta man converter */
#define RMAN_PATH "/usr/bin/rman"
/* legal characters (alphanumeric is always allowed) for section and 
   man page name */
#define LEGAL_SECTION "-._+:"
#define LEGAL_NAME "-._+:"
#define MAX_NAME 60
#define MAX_SECTION 10

static int debug = 0;


static int
http_error(int error)
{
    char *p;

    switch (error) {
    case 400:
	p = "Bad Request";
	break;
    case 404:
	p = "Not Found";
	break;
    default:
	p = "Error";
    }
    printf("HTTP/1.0 %d %s\r\n", error, p);
    printf("MIME-version: 1.0\r\n");
    printf("Content-Type: text/html\r\n\r\n");
    printf("<head><title>%d %s</title></head>\r\n", error, p);
    printf("<body><h1>%d %s</h1></body>\r\n", error, p);
    return 0;
}

/*
 * checks s1 for correctness, that is s1 is alphanumeric
 * and may contain chars from sc
 * and should not be longer than maxl
 * returns 0 on success, 1 on failure
 */
static int
str_check (const char *s1, const char *sc, int maxl)
{
  if (*s1 == '-') /* s1's first character may not be "-", marc@suse.de */
      return 1;
  while (maxl-- 
	 && *s1 
	 && (isalnum ((unsigned char) *s1) || strchr (sc, *s1)))
    ++s1;
  return (maxl < 0 || *s1 != '\0');
}

static char tmp_file[64];
static int fd = -1;
static sighandler_t sighandler_hup;
static sighandler_t sighandler_int;
static sighandler_t sighandler_term;

static void sighandler(const int signo)
{
    sighandler_t ret;

    if (fd > 2) {
	close(fd);
	unlink(tmp_file);
	fd = -1;
    }

    if ((ret = signal(signo, SIG_DFL)) == SIG_ERR) {
	_exit (2);  /* signal handlers shouldn't exit */
    }
    kill (getpid(), signo);
    abort();
}

static int sethandler()
{
    if ((sighandler_hup  = signal(SIGHUP,  sighandler)) == SIG_ERR ||
	(sighandler_int  = signal(SIGINT,  sighandler)) == SIG_ERR ||
	(sighandler_term = signal(SIGTERM, sighandler)) == SIG_ERR)
	return -1;
    return 0;
}

static int unsethandler()
{
    sighandler_t ret;
    if ((ret = signal(SIGHUP,  sighandler_hup))  == SIG_ERR ||
	(ret = signal(SIGINT,  sighandler_int))  == SIG_ERR ||
	(ret = signal(SIGTERM, sighandler_term)) == SIG_ERR)
	return -1;
    return 0;
}

static int
http_rman(char *url)
{
    char *pName;
    char *pSection;
    char buf[1024];
    int status;
    char* tok1;
    char* s;

    if (debug)
      syslog (LOG_NOTICE, "url(%s)", url);
    pName = NULL;
    pSection = NULL;
    /* parse URL: should be /man?p=command[&s=section] */
    tok1 = strtok( url, "=" ); /* man?p= */
    if (!tok1)
      return -1;
    pName = strtok( NULL, "=" ); /* command[&s] */
    if (!pName)
      return -1;
    pSection = strtok( NULL, "=" ); /* section */
    /* strip & from command */
    if (s = strrchr (pName, '&')) 
      *s = '\0';
    if(!pSection) 
      pSection = "";

    if (debug)
      syslog (LOG_NOTICE, "found name(%s)sec(%s)", pName, pSection);

    /* tmg@suse.de: check section and name for legal name */
    if (str_check (pSection, LEGAL_SECTION, MAX_SECTION)) {
//      if (debug) 
	syslog (LOG_NOTICE, "illegal character in Section(%s)", pSection);
      return -1;
    }
    if (str_check (pName, LEGAL_NAME, MAX_NAME)) {
//      if (debug)
	syslog (LOG_NOTICE, "illegal character in Name(%s)", pName);
      return -1;
    }

    /* generate an (hopefully) unused filename */
    sprintf( tmp_file, "/tmp/susehelp%ld.XXXXXX", (unsigned long) getpid());
    errno = 0;
    fd = mkstemp(tmp_file);
    if (errno) {
	fprintf(stderr, "mkstemp(%s) failed: %s\n", tmp_file, sys_errlist[errno]);
	exit(1);
    } 
    sethandler();
    /* check for buffer overflow! tmg@suse.de, we have plenty of room */
    unsetenv("MANPATH");
    if (snprintf( buf, sizeof (buf), "%s -P /bin/cat %s %s &>%s ",
		  MAN_PATH, pSection, pName, tmp_file) < 0) {
//      if (debug)
	syslog (LOG_NOTICE, "buffer overflow in snprintf name(%s)(%s)",
		pName, pSection);
      return -1;
    }
    if (debug)
      syslog (LOG_NOTICE, "system(%s)", buf);
    status = WEXITSTATUS(system( buf ));
    if (debug && status != 0)
      syslog (LOG_NOTICE, "system rc = %d %m", status);

    printf("HTTP/1.0 200 OK\r\n" );
    printf("MIME-version: 1.0\r\n");
    printf("Content-Type: text/html\r\n\r\n");
    fflush(stdout);
    if( status == 0){
      if (snprintf(buf, sizeof (buf), "/bin/cat %s | %s -r 'man?=%%s&=%%s' -l 'man %%s (%%s)' -n %s %s %s -f HTML",
		   tmp_file, RMAN_PATH, pName, *pSection ? "-s" : "", pSection) < 0) {
//	if (debug)
	  syslog (LOG_NOTICE, "buffer overflow in rman command name(%s)section(%s)", 
		  pName, pSection);
	return -1;
      }

      if (debug)
	syslog (LOG_NOTICE, "system(%s)", buf);
      system( buf );
      fflush(stdout);
      }
    else
        {

        printf( "<html>\n" );
        printf( "<head><title>RosettaMan Gateway</title></head>\n" );
        printf( "<h1>Manual Page not found</h1>\n" );
        printf( "The requested page about <STRONG>\n" );
        fflush(stdout);
	printf ("%s(%s)\n", pName, pSection);
        fflush(stdout);

        printf( "</STRONG> could not be found.<P>\n" );
        printf( "This may have several reasones:\n<UL>\n" );
        printf( "<LI>The requested manual page is not installed.<BR>\n" );
        printf( "Install the corresponding package (man...) of the ap- or d-series,\n" );
        printf( " respectively the corresponding application from the z-series.\n" );
        printf( "<LI>You followed a link from another manual-page. Some\n" );
        printf( "manual pages refer to non-existing pages. This cannot be totally avoided,\n" );
        printf( "because the manual pages are maintained by hundreds of independent developers :-(\n" );
        printf( "</UL>\n" );
        printf( "Go back using the <STRONG>BACK</STRONG>-funktion of your 3W-Browser.\n" );
        printf( "</HTML>\n" );
        fflush(stdout);
        }

    printf( "\n" );
    if (fd > 2) {
	close(fd);
	unlink (tmp_file);
	fd = -1;
    }
    unsethandler();
    return 0;
}


int
main(int ac, char **av)
{
    char buf[200];
    char url[200];
    int status;
    int i;
    char *sFrontpage = "/usr/doc/susehilf/man/index.html";

    /* simplistic option parsing tmg@suse.de */
    for (i = 1; i < ac; ++i)
      if (!strcmp (av[i], "-d"))
	debug = 1;

    openlog ("httpd-rman", LOG_CONS|LOG_PID, LOG_DAEMON);

/* just read in one line from stdin and make sure it is a GET
   command */

    if (fgets(buf, sizeof (buf), stdin)) {

	/* read rest of command (just for the sake of it) */
	while (fgets (url, sizeof (url), stdin) && url[0] != '\r')
	    ;

	/* command should be GET <url> [HTTP/1.0] */
	if (sscanf(buf, "GET %199s\n", url) == 1) {

	    status  = http_rman(url);

	    if (status < 0) 
                {
		  FILE *fp = fopen (sFrontpage, "r");
		  if (fp) {
		    int c;
		    while ((c = getc (fp)) != EOF) {
		      putc (c, stdout);
		    }
		    fclose (fp);
		    status = 0;
		  }
		}
	    if (status < 0)
	      http_error(404);

	} else

	    http_error(400);
    }

    fflush(stdout);
    closelog ();
    exit(0);
}
